..
    Warning: Do not edit this file. It is automatically generated from the
    software project's code and your changes will be overwritten.

    The tool to generate this file lives in openstack-doc-tools repository.

    Please make any changes needed in the code, then run the
    autogenerate-config-doc tool from the openstack-doc-tools repository, or
    ask for help on the documentation mailing list, IRC channel or meeting.

.. _manila-quobyte:

.. list-table:: Description of Quobyte share driver configuration options
   :header-rows: 1
   :class: config-ref-table

   * - Configuration option = Default value
     - Description
   * - **[DEFAULT]**
     -
   * - ``quobyte_api_ca`` = ``None``
     - (String) The X.509 CA file to verify the server cert.
   * - ``quobyte_api_password`` = ``quobyte``
     - (String) Password for Quobyte API server
   * - ``quobyte_api_url`` = ``None``
     - (String) URL of the Quobyte API server (http or https)
   * - ``quobyte_api_username`` = ``admin``
     - (String) Username for Quobyte API server.
   * - ``quobyte_default_volume_group`` = ``root``
     - (String) Default owning group for new volumes.
   * - ``quobyte_default_volume_user`` = ``root``
     - (String) Default owning user for new volumes.
   * - ``quobyte_delete_shares`` = ``False``
     - (Boolean) Actually deletes shares (vs. unexport)
   * - ``quobyte_volume_configuration`` = ``BASE``
     - (String) Name of volume configuration used for new shares.
