..
    Warning: Do not edit this file. It is automatically generated from the
    software project's code and your changes will be overwritten.

    The tool to generate this file lives in openstack-doc-tools repository.

    Please make any changes needed in the code, then run the
    autogenerate-config-doc tool from the openstack-doc-tools repository, or
    ask for help on the documentation mailing list, IRC channel or meeting.

.. _manila-compute:

.. list-table:: Description of Compute configuration options
   :header-rows: 1
   :class: config-ref-table

   * - Configuration option = Default value
     - Description
   * - **[DEFAULT]**
     -
   * - ``nova_admin_auth_url`` = ``http://localhost:5000/v2.0``
     - (String) DEPRECATED: Identity service URL. This option isn't used any longer. Please use [nova] url instead.
   * - ``nova_admin_password`` = ``None``
     - (String) DEPRECATED: Nova admin password. This option isn't used any longer. Please use [nova] password instead.
   * - ``nova_admin_tenant_name`` = ``service``
     - (String) DEPRECATED: Nova admin tenant name. This option isn't used any longer. Please use [nova] tenant instead.
   * - ``nova_admin_username`` = ``nova``
     - (String) DEPRECATED: Nova admin username. This option isn't used any longer. Please use [nova] username instead.
   * - ``nova_catalog_admin_info`` = ``compute:nova:adminURL``
     - (String) DEPRECATED: Same as nova_catalog_info, but for admin endpoint. This option isn't used any longer.
   * - ``nova_catalog_info`` = ``compute:nova:publicURL``
     - (String) DEPRECATED: Info to match when looking for nova in the service catalog. Format is separated values of the form: <service_type>:<service_name>:<endpoint_type> This option isn't used any longer.
   * - ``os_region_name`` = ``None``
     - (String) Region name of this node.
