..
    Warning: Do not edit this file. It is automatically generated from the
    software project's code and your changes will be overwritten.

    The tool to generate this file lives in openstack-doc-tools repository.

    Please make any changes needed in the code, then run the
    autogenerate-config-doc tool from the openstack-doc-tools repository, or
    ask for help on the documentation mailing list, IRC channel or meeting.

.. _manila-common:

.. list-table:: Description of Common configuration options
   :header-rows: 1
   :class: config-ref-table

   * - Configuration option = Default value
     - Description
   * - **[DEFAULT]**
     -
   * - ``check_hash`` = ``False``
     - (Boolean) Chooses whether hash of each file should be checked on data copying.
   * - ``client_socket_timeout`` = ``900``
     - (Integer) Timeout for client connections socket operations. If an incoming connection is idle for this number of seconds it will be closed. A value of '0' means wait forever.
   * - ``compute_api_class`` = ``manila.compute.nova.API``
     - (String) The full class name of the Compute API class to use.
   * - ``data_access_wait_access_rules_timeout`` = ``180``
     - (Integer) Time to wait for access rules to be allowed/denied on backends when migrating a share (seconds).
   * - ``data_manager`` = ``manila.data.manager.DataManager``
     - (String) Full class name for the data manager.
   * - ``data_node_access_admin_user`` = ``None``
     - (String) The admin user name registered in the security service in order to allow access to user authentication-based shares.
   * - ``data_node_access_cert`` = ``None``
     - (String) The certificate installed in the data node in order to allow access to certificate authentication-based shares.
   * - ``data_node_access_ip`` = ``None``
     - (String) The IP of the node interface connected to the admin network. Used for allowing access to the mounting shares.
   * - ``data_node_mount_options`` = ``{}``
     - (Dict) Mount options to be included in the mount command for share protocols. Use dictionary format, example: {'nfs': '-o nfsvers=3', 'cifs': '-o user=foo,pass=bar'}
   * - ``data_topic`` = ``manila-data``
     - (String) The topic data nodes listen on.
   * - ``enable_new_services`` = ``True``
     - (Boolean) Services to be added to the available pool on create.
   * - ``fatal_exception_format_errors`` = ``False``
     - (Boolean) Whether to make exception message format errors fatal.
   * - ``filter_function`` = ``None``
     - (String) String representation for an equation that will be used to filter hosts.
   * - ``host`` = ``<your_hostname>``
     - (String) Name of this node. This can be an opaque identifier. It is not necessarily a hostname, FQDN, or IP address.
   * - ``max_over_subscription_ratio`` = ``20.0``
     - (Floating point) Float representation of the over subscription ratio when thin provisioning is involved. Default ratio is 20.0, meaning provisioned capacity can be 20 times the total physical capacity. If the ratio is 10.5, it means provisioned capacity can be 10.5 times the total physical capacity. A ratio of 1.0 means provisioned capacity cannot exceed the total physical capacity. A ratio lower than 1.0 is invalid.
   * - ``memcached_servers`` = ``None``
     - (List) Memcached servers or None for in process cache.
   * - ``monkey_patch`` = ``False``
     - (Boolean) Whether to log monkey patching.
   * - ``monkey_patch_modules`` =
     - (List) List of modules or decorators to monkey patch.
   * - ``mount_tmp_location`` = ``/tmp/``
     - (String) Temporary path to create and mount shares during migration.
   * - ``my_ip`` = ``<your_ip>``
     - (String) IP address of this host.
   * - ``num_shell_tries`` = ``3``
     - (Integer) Number of times to attempt to run flakey shell commands.
   * - ``periodic_fuzzy_delay`` = ``60``
     - (Integer) Range of seconds to randomly delay when starting the periodic task scheduler to reduce stampeding. (Disable by setting to 0)
   * - ``periodic_hooks_interval`` = ``300.0``
     - (Floating point) Interval in seconds between execution of periodic hooks. Used when option 'enable_periodic_hooks' is set to True. Default is 300.
   * - ``periodic_interval`` = ``60``
     - (Integer) Seconds between running periodic tasks.
   * - ``replica_state_update_interval`` = ``300``
     - (Integer) This value, specified in seconds, determines how often the share manager will poll for the health (replica_state) of each replica instance.
   * - ``replication_domain`` = ``None``
     - (String) A string specifying the replication domain that the backend belongs to. This option needs to be specified the same in the configuration sections of all backends that support replication between each other. If this option is not specified in the group, it means that replication is not enabled on the backend.
   * - ``report_interval`` = ``10``
     - (Integer) Seconds between nodes reporting state to datastore.
   * - ``reserved_share_percentage`` = ``0``
     - (Integer) The percentage of backend capacity reserved.
   * - ``rootwrap_config`` = ``None``
     - (String) Path to the rootwrap configuration file to use for running commands as root.
   * - ``service_down_time`` = ``60``
     - (Integer) Maximum time since last check-in for up service.
   * - ``smb_template_config_path`` = ``$state_path/smb.conf``
     - (String) Path to smb config.
   * - ``sql_idle_timeout`` = ``3600``
     - (Integer) Timeout before idle SQL connections are reaped.
   * - ``sql_max_retries`` = ``10``
     - (Integer) Maximum database connection retries during startup. (setting -1 implies an infinite retry count).
   * - ``sql_retry_interval`` = ``10``
     - (Integer) Interval between retries of opening a SQL connection.
   * - ``sqlite_clean_db`` = ``clean.sqlite``
     - (String) File name of clean sqlite database.
   * - ``sqlite_db`` = ``manila.sqlite``
     - (String) The filename to use with sqlite.
   * - ``sqlite_synchronous`` = ``True``
     - (Boolean) If passed, use synchronous mode for sqlite.
   * - ``state_path`` = ``/var/lib/manila``
     - (String) Top-level directory for maintaining manila's state.
   * - ``storage_availability_zone`` = ``nova``
     - (String) Availability zone of this node.
   * - ``tcp_keepalive`` = ``True``
     - (Boolean) Sets the value of TCP_KEEPALIVE (True/False) for each server socket.
   * - ``tcp_keepalive_count`` = ``None``
     - (Integer) Sets the value of TCP_KEEPCNT for each server socket. Not supported on OS X.
   * - ``tcp_keepalive_interval`` = ``None``
     - (Integer) Sets the value of TCP_KEEPINTVL in seconds for each server socket. Not supported on OS X.
   * - ``tcp_keepidle`` = ``600``
     - (Integer) Sets the value of TCP_KEEPIDLE in seconds for each server socket. Not supported on OS X.
   * - ``until_refresh`` = ``0``
     - (Integer) Count of reservations until usage is refreshed.
   * - ``use_forwarded_for`` = ``False``
     - (Boolean) Treat X-Forwarded-For as the canonical remote address. Only enable this if you have a sanitizing proxy.
   * - ``wsgi_keep_alive`` = ``True``
     - (Boolean) If False, closes the client socket connection explicitly. Setting it to True to maintain backward compatibility. Recommended setting is set it to False.
   * - **[coordination]**
     -
   * - ``backend_url`` = ``file://$state_path``
     - (String) The back end URL to use for distributed coordination.
   * - **[healthcheck]**
     -
   * - ``backends`` =
     - (List) Additional backends that can perform health checks and report that information back as part of a request.
   * - ``detailed`` = ``False``
     - (Boolean) Show more detailed information as part of the response
   * - ``disable_by_file_path`` = ``None``
     - (String) Check the presence of a file to determine if an application is running on a port. Used by DisableByFileHealthcheck plugin.
   * - ``disable_by_file_paths`` =
     - (List) Check the presence of a file based on a port to determine if an application is running on a port. Expects a "port:path" list of strings. Used by DisableByFilesPortsHealthcheck plugin.
   * - ``path`` = ``/healthcheck``
     - (String) DEPRECATED: The path to respond to healtcheck requests on.
