..
    Warning: Do not edit this file. It is automatically generated from the
    software project's code and your changes will be overwritten.

    The tool to generate this file lives in openstack-doc-tools repository.

    Please make any changes needed in the code, then run the
    autogenerate-config-doc tool from the openstack-doc-tools repository, or
    ask for help on the documentation mailing list, IRC channel or meeting.

.. _manila-scheduler:

.. list-table:: Description of Scheduler configuration options
   :header-rows: 1
   :class: config-ref-table

   * - Configuration option = Default value
     - Description
   * - **[DEFAULT]**
     -
   * - ``capacity_weight_multiplier`` = ``1.0``
     - (Floating point) Multiplier used for weighing share capacity. Negative numbers mean to stack vs spread.
   * - ``pool_weight_multiplier`` = ``1.0``
     - (Floating point) Multiplier used for weighing pools which have existing share servers. Negative numbers mean to spread vs stack.
   * - ``scheduler_default_filters`` = ``AvailabilityZoneFilter, CapacityFilter, CapabilitiesFilter, DriverFilter, ShareReplicationFilter``
     - (List) Which filter class names to use for filtering hosts when not specified in the request.
   * - ``scheduler_default_weighers`` = ``CapacityWeigher, GoodnessWeigher``
     - (List) Which weigher class names to use for weighing hosts.
   * - ``scheduler_driver`` = ``manila.scheduler.drivers.filter.FilterScheduler``
     - (String) Default scheduler driver to use.
   * - ``scheduler_host_manager`` = ``manila.scheduler.host_manager.HostManager``
     - (String) The scheduler host manager class to use.
   * - ``scheduler_json_config_location`` =
     - (String) Absolute path to scheduler configuration JSON file.
   * - ``scheduler_manager`` = ``manila.scheduler.manager.SchedulerManager``
     - (String) Full class name for the scheduler manager.
   * - ``scheduler_max_attempts`` = ``3``
     - (Integer) Maximum number of attempts to schedule a share.
   * - ``scheduler_topic`` = ``manila-scheduler``
     - (String) The topic scheduler nodes listen on.
