..
    Warning: Do not edit this file. It is automatically generated from the
    software project's code and your changes will be overwritten.

    The tool to generate this file lives in openstack-doc-tools repository.

    Please make any changes needed in the code, then run the
    autogenerate-config-doc tool from the openstack-doc-tools repository, or
    ask for help on the documentation mailing list, IRC channel or meeting.

.. _manila-share:

.. list-table:: Description of Share configuration options
   :header-rows: 1
   :class: config-ref-table

   * - Configuration option = Default value
     - Description
   * - **[DEFAULT]**
     -
   * - ``automatic_share_server_cleanup`` = ``True``
     - (Boolean) If set to True, then Manila will delete all share servers which were unused more than specified time .If set to False - automatic deletion of share servers will be disabled.
   * - ``backlog`` = ``4096``
     - (Integer) Number of backlog requests to configure the socket with.
   * - ``default_share_group_type`` = ``None``
     - (String) Default share group type to use.
   * - ``default_share_type`` = ``None``
     - (String) Default share type to use.
   * - ``delete_share_server_with_last_share`` = ``False``
     - (Boolean) Whether share servers will be deleted on deletion of the last share.
   * - ``driver_handles_share_servers`` = ``None``
     - (Boolean) There are two possible approaches for share drivers in Manila. First is when share driver is able to handle share-servers and second when not. Drivers can support either both or only one of these approaches. So, set this opt to True if share driver is able to handle share servers and it is desired mode else set False. It is set to None by default to make this choice intentional.
   * - ``enable_periodic_hooks`` = ``False``
     - (Boolean) Whether to enable periodic hooks or not.
   * - ``enable_post_hooks`` = ``False``
     - (Boolean) Whether to enable post hooks or not.
   * - ``enable_pre_hooks`` = ``False``
     - (Boolean) Whether to enable pre hooks or not.
   * - ``enabled_share_backends`` = ``None``
     - (List) A list of share backend names to use. These backend names should be backed by a unique [CONFIG] group with its options.
   * - ``enabled_share_protocols`` = ``NFS, CIFS``
     - (List) Specify list of protocols to be allowed for share creation. Available values are '('NFS', 'CIFS', 'GLUSTERFS', 'HDFS', 'CEPHFS', 'MAPRFS')'
   * - ``executor_thread_pool_size`` = ``64``
     - (Integer) Size of executor thread pool.
   * - ``hook_drivers`` =
     - (List) Driver(s) to perform some additional actions before and after share driver actions and on a periodic basis. Default is [].
   * - ``migration_create_delete_share_timeout`` = ``300``
     - (Integer) Timeout for creating and deleting share instances when performing share migration (seconds).
   * - ``migration_driver_continue_update_interval`` = ``60``
     - (Integer) This value, specified in seconds, determines how often the share manager will poll the driver to perform the next step of migration in the storage backend, for a migrating share.
   * - ``migration_ignore_files`` = ``lost+found``
     - (List) List of files and folders to be ignored when migrating shares. Items should be names (not including any path).
   * - ``migration_readonly_rules_support`` = ``True``
     - (Boolean) DEPRECATED: Specify whether read only access rule mode is supported in this backend. Obsolete. All drivers are now required to support read-only access rules.
   * - ``migration_wait_access_rules_timeout`` = ``180``
     - (Integer) Time to wait for access rules to be allowed/denied on backends when migrating shares using generic approach (seconds).
   * - ``network_config_group`` = ``None``
     - (String) Name of the configuration group in the Manila conf file to look for network config options.If not set, the share backend's config group will be used.If an option is not found within provided group, then'DEFAULT' group will be used for search of option.
   * - ``share_manager`` = ``manila.share.manager.ShareManager``
     - (String) Full class name for the share manager.
   * - ``share_name_template`` = ``share-%s``
     - (String) Template string to be used to generate share names.
   * - ``share_snapshot_name_template`` = ``share-snapshot-%s``
     - (String) Template string to be used to generate share snapshot names.
   * - ``share_topic`` = ``manila-share``
     - (String) The topic share nodes listen on.
   * - ``suppress_post_hooks_errors`` = ``False``
     - (Boolean) Whether to suppress post hook errors (allow driver's results to pass through) or not.
   * - ``suppress_pre_hooks_errors`` = ``False``
     - (Boolean) Whether to suppress pre hook errors (allow driver perform actions) or not.
   * - ``unmanage_remove_access_rules`` = ``False``
     - (Boolean) If set to True, then manila will deny access and remove all access rules on share unmanage.If set to False - nothing will be changed.
   * - ``unused_share_server_cleanup_interval`` = ``10``
     - (Integer) Unallocated share servers reclamation time interval (minutes). Minimum value is 10 minutes, maximum is 60 minutes. The reclamation function is run every 10 minutes and delete share servers which were unused more than unused_share_server_cleanup_interval option defines. This value reflects the shortest time Manila will wait for a share server to go unutilized before deleting it.
   * - ``use_scheduler_creating_share_from_snapshot`` = ``False``
     - (Boolean) If set to False, then share creation from snapshot will be performed on the same host. If set to True, then scheduling step will be used.
