..
    Warning: Do not edit this file. It is automatically generated from the
    software project's code and your changes will be overwritten.

    The tool to generate this file lives in openstack-doc-tools repository.

    Please make any changes needed in the code, then run the
    autogenerate-config-doc tool from the openstack-doc-tools repository, or
    ask for help on the documentation mailing list, IRC channel or meeting.

.. _manila-cephfs:

.. list-table:: Description of CephFS share driver configuration options
   :header-rows: 1
   :class: config-ref-table

   * - Configuration option = Default value
     - Description
   * - **[DEFAULT]**
     -
   * - ``cephfs_auth_id`` = ``manila``
     - (String) The name of the ceph auth identity to use.
   * - ``cephfs_cluster_name`` = ``None``
     - (String) The name of the cluster in use, if it is not the default ('ceph').
   * - ``cephfs_conf_path`` =
     - (String) Fully qualified path to the ceph.conf file.
   * - ``cephfs_enable_snapshots`` = ``False``
     - (Boolean) Whether to enable snapshots in this driver.
