..
    Warning: Do not edit this file. It is automatically generated from the
    software project's code and your changes will be overwritten.

    The tool to generate this file lives in openstack-doc-tools repository.

    Please make any changes needed in the code, then run the
    autogenerate-config-doc tool from the openstack-doc-tools repository, or
    ask for help on the documentation mailing list, IRC channel or meeting.

.. _manila-maprfs:

.. list-table:: Description of MapRFS share driver configuration options
   :header-rows: 1
   :class: config-ref-table

   * - Configuration option = Default value
     - Description
   * - **[DEFAULT]**
     -
   * - ``maprfs_base_volume_dir`` = ``/``
     - (String) Path in MapRFS where share volumes must be created.
   * - ``maprfs_cldb_ip`` = ``None``
     - (List) The list of IPs or hostnames of CLDB nodes.
   * - ``maprfs_clinode_ip`` = ``None``
     - (List) The list of IPs or hostnames of nodes where mapr-core is installed.
   * - ``maprfs_rename_managed_volume`` = ``True``
     - (Boolean) Specify whether existing volume should be renamed when start managing.
   * - ``maprfs_ssh_name`` = ``mapr``
     - (String) Cluster admin user ssh login name.
   * - ``maprfs_ssh_port`` = ``22``
     - (Port number) CLDB node SSH port.
   * - ``maprfs_ssh_private_key`` = ``None``
     - (String) Path to SSH private key for login.
   * - ``maprfs_ssh_pw`` = ``None``
     - (String) Cluster node SSH login password, This parameter is not necessary, if 'maprfs_ssh_private_key' is configured.
   * - ``maprfs_zookeeper_ip`` = ``None``
     - (List) The list of IPs or hostnames of ZooKeeper nodes.
