..
    Warning: Do not edit this file. It is automatically generated from the
    software project's code and your changes will be overwritten.

    The tool to generate this file lives in openstack-doc-tools repository.

    Please make any changes needed in the code, then run the
    autogenerate-config-doc tool from the openstack-doc-tools repository, or
    ask for help on the documentation mailing list, IRC channel or meeting.

.. _manila-zfssa:

.. list-table:: Description of ZFSSA share driver configuration options
   :header-rows: 1
   :class: config-ref-table

   * - Configuration option = Default value
     - Description
   * - **[DEFAULT]**
     -
   * - ``zfssa_auth_password`` = ``None``
     - (String) ZFSSA management authorized userpassword.
   * - ``zfssa_auth_user`` = ``None``
     - (String) ZFSSA management authorized username.
   * - ``zfssa_data_ip`` = ``None``
     - (String) IP address for data.
   * - ``zfssa_host`` = ``None``
     - (String) ZFSSA management IP address.
   * - ``zfssa_manage_policy`` = ``loose``
     - (String) Driver policy for share manage. A strict policy checks for a schema named manila_managed, and makes sure its value is true. A loose policy does not check for the schema.
   * - ``zfssa_nas_checksum`` = ``fletcher4``
     - (String) Controls checksum used for data blocks.
   * - ``zfssa_nas_compression`` = ``off``
     - (String) Data compression-off, lzjb, gzip-2, gzip, gzip-9.
   * - ``zfssa_nas_logbias`` = ``latency``
     - (String) Controls behavior when servicing synchronous writes.
   * - ``zfssa_nas_mountpoint`` =
     - (String) Location of project in ZFS/SA.
   * - ``zfssa_nas_quota_snap`` = ``true``
     - (String) Controls whether a share quota includes snapshot.
   * - ``zfssa_nas_rstchown`` = ``true``
     - (String) Controls whether file ownership can be changed.
   * - ``zfssa_nas_vscan`` = ``false``
     - (String) Controls whether the share is scanned for viruses.
   * - ``zfssa_pool`` = ``None``
     - (String) ZFSSA storage pool name.
   * - ``zfssa_project`` = ``None``
     - (String) ZFSSA project name.
   * - ``zfssa_rest_timeout`` = ``None``
     - (String) REST connection timeout (in seconds).
