..
    Warning: Do not edit this file. It is automatically generated from the
    software project's code and your changes will be overwritten.

    The tool to generate this file lives in openstack-doc-tools repository.

    Please make any changes needed in the code, then run the
    autogenerate-config-doc tool from the openstack-doc-tools repository, or
    ask for help on the documentation mailing list, IRC channel or meeting.

.. _manila-hpe3par:

.. list-table:: Description of HPE 3PAR share driver configuration options
   :header-rows: 1
   :class: config-ref-table

   * - Configuration option = Default value
     - Description
   * - **[DEFAULT]**
     -
   * - ``hpe3par_api_url`` =
     - (String) 3PAR WSAPI Server Url like https://<3par ip>:8080/api/v1
   * - ``hpe3par_cifs_admin_access_domain`` = ``LOCAL_CLUSTER``
     - (String) File system domain for the CIFS admin user.
   * - ``hpe3par_cifs_admin_access_password`` =
     - (String) File system admin password for CIFS.
   * - ``hpe3par_cifs_admin_access_username`` =
     - (String) File system admin user name for CIFS.
   * - ``hpe3par_debug`` = ``False``
     - (Boolean) Enable HTTP debugging to 3PAR
   * - ``hpe3par_fpg`` = ``None``
     - (Unknown) The File Provisioning Group (FPG) to use
   * - ``hpe3par_fstore_per_share`` = ``False``
     - (Boolean) Use one filestore per share
   * - ``hpe3par_password`` =
     - (String) 3PAR password for the user specified in hpe3par_username
   * - ``hpe3par_require_cifs_ip`` = ``False``
     - (Boolean) Require IP access rules for CIFS (in addition to user)
   * - ``hpe3par_san_ip`` =
     - (String) IP address of SAN controller
   * - ``hpe3par_san_login`` =
     - (String) Username for SAN controller
   * - ``hpe3par_san_password`` =
     - (String) Password for SAN controller
   * - ``hpe3par_san_ssh_port`` = ``22``
     - (Port number) SSH port to use with SAN
   * - ``hpe3par_share_mount_path`` = ``/mnt/``
     - (String) The path where shares will be mounted when deleting nested file trees.
   * - ``hpe3par_username`` =
     - (String) 3PAR username with the 'edit' role
