..
    Warning: Do not edit this file. It is automatically generated from the
    software project's code and your changes will be overwritten.

    The tool to generate this file lives in openstack-doc-tools repository.

    Please make any changes needed in the code, then run the
    autogenerate-config-doc tool from the openstack-doc-tools repository, or
    ask for help on the documentation mailing list, IRC channel or meeting.

.. _manila-netapp:

.. list-table:: Description of NetApp share driver configuration options
   :header-rows: 1
   :class: config-ref-table

   * - Configuration option = Default value
     - Description
   * - **[DEFAULT]**
     -
   * - ``netapp_aggregate_name_search_pattern`` = ``(.*)``
     - (String) Pattern for searching available aggregates for provisioning.
   * - ``netapp_enabled_share_protocols`` = ``nfs3, nfs4.0``
     - (List) The NFS protocol versions that will be enabled. Supported values include nfs3, nfs4.0, nfs4.1. This option only applies when the option driver_handles_share_servers is set to True.
   * - ``netapp_lif_name_template`` = ``os_%(net_allocation_id)s``
     - (String) Logical interface (LIF) name template
   * - ``netapp_login`` = ``None``
     - (String) Administrative user account name used to access the storage system.
   * - ``netapp_password`` = ``None``
     - (String) Password for the administrative user account specified in the netapp_login option.
   * - ``netapp_port_name_search_pattern`` = ``(.*)``
     - (String) Pattern for overriding the selection of network ports on which to create Vserver LIFs.
   * - ``netapp_root_volume`` = ``root``
     - (String) Root volume name.
   * - ``netapp_root_volume_aggregate`` = ``None``
     - (String) Name of aggregate to create Vserver root volumes on. This option only applies when the option driver_handles_share_servers is set to True.
   * - ``netapp_server_hostname`` = ``None``
     - (String) The hostname (or IP address) for the storage system.
   * - ``netapp_server_port`` = ``None``
     - (Port number) The TCP port to use for communication with the storage system or proxy server. If not specified, Data ONTAP drivers will use 80 for HTTP and 443 for HTTPS.
   * - ``netapp_snapmirror_quiesce_timeout`` = ``3600``
     - (Integer) The maximum time in seconds to wait for existing snapmirror transfers to complete before aborting when promoting a replica.
   * - ``netapp_storage_family`` = ``ontap_cluster``
     - (String) The storage family type used on the storage system; valid values include ontap_cluster for using clustered Data ONTAP.
   * - ``netapp_trace_flags`` = ``None``
     - (String) Comma-separated list of options that control which trace info is written to the debug logs. Values include method and api.
   * - ``netapp_transport_type`` = ``http``
     - (String) The transport protocol used when communicating with the storage system or proxy server. Valid values are http or https.
   * - ``netapp_volume_move_cutover_timeout`` = ``3600``
     - (Integer) The maximum time in seconds to wait for the completion of a volume move operation after the cutover was triggered.
   * - ``netapp_volume_name_template`` = ``share_%(share_id)s``
     - (String) NetApp volume name template.
   * - ``netapp_volume_snapshot_reserve_percent`` = ``5``
     - (Integer) The percentage of share space set aside as reserve for snapshot usage; valid values range from 0 to 90.
   * - ``netapp_vserver_name_template`` = ``os_%s``
     - (String) Name template to use for new Vserver. When using CIFS protocol make sure to not configure characters illegal in DNS hostnames.
