# Copyright 2015 Mirantis Inc.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""add_nova_net_id_column_to_share_networks

Revision ID: 17115072e1c3
Revises: 38e632621e5a
Create Date: 2015-02-05 18:07:19.062995

"""

# revision identifiers, used by Alembic.
revision = '17115072e1c3'
down_revision = '38e632621e5a'

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.add_column(
        'share_networks',
        sa.Column('nova_net_id', sa.String(36), nullable=True))


def downgrade():
    op.drop_column('share_networks', 'nova_net_id')
