# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""add_gateway_to_network_allocations_table

Revision ID: fdfb668d19e1
Revises: 221a83cfd85b
Create Date: 2016-04-19 10:07:16.224806

"""

# revision identifiers, used by Alembic.
revision = 'fdfb668d19e1'
down_revision = '221a83cfd85b'

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.add_column(
        'network_allocations',
        sa.Column('gateway', sa.String(64), nullable=True))
    op.add_column(
        'share_networks',
        sa.Column('gateway', sa.String(64), nullable=True))


def downgrade():
    op.drop_column('network_allocations', 'gateway')
    op.drop_column('share_networks', 'gateway')
